use sprinklers_core::model::{ProgramId, ProgramRef, ProgramUpdateData, Programs};
use thiserror::Error;
use tokio::sync::{mpsc, oneshot, watch};

#[derive(Debug)]
pub enum Request {
    UpdateProgram {
        id: ProgramId,
        update: ProgramUpdateData,
        resp_tx: oneshot::Sender<Result<ProgramRef>>,
    },
}

#[derive(Clone)]
pub struct StateManager {
    request_tx: mpsc::Sender<Request>,
    programs_watch: watch::Receiver<Programs>,
}

#[derive(Debug, Error)]
pub enum StateError {
    #[error("no such program: {0}")]
    NoSuchProgram(ProgramId),
    #[error("internal error: {0}")]
    Other(
        #[from]
        #[source]
        eyre::Report,
    ),
}

pub type Result<T, E = StateError> = std::result::Result<T, E>;

impl StateManager {
    pub fn new(
        request_tx: mpsc::Sender<Request>,
        programs_watch: watch::Receiver<Programs>,
    ) -> Self {
        Self {
            request_tx,
            programs_watch,
        }
    }

    pub async fn update_program(
        &mut self,
        id: ProgramId,
        update: ProgramUpdateData,
    ) -> Result<ProgramRef> {
        let (resp_tx, resp_rx) = oneshot::channel();
        self.request_tx
            .send(Request::UpdateProgram {
                id,
                update,
                resp_tx,
            })
            .await.map_err(eyre::Report::from)?;
        resp_rx.await.map_err(eyre::Report::from)?
    }

    pub fn get_programs(&self) -> watch::Receiver<Programs> {
        self.programs_watch.clone()
    }
}
