use std::iter::repeat_with;
use std::sync::atomic::{AtomicBool, Ordering};
use tracing::debug;

pub type SecId = u32;

pub trait SectionInterface: Send {
    fn num_sections(&self) -> SecId;
    fn set_section_state(&self, id: SecId, running: bool);
    fn get_section_state(&self, id: SecId) -> bool;
}

pub struct MockSectionInterface {
    states: Vec<AtomicBool>,
}

impl MockSectionInterface {
    #[allow(dead_code)]
    pub fn new(num_sections: SecId) -> Self {
        Self {
            states: repeat_with(|| AtomicBool::new(false))
                .take(num_sections as usize)
                .collect(),
        }
    }
}

impl SectionInterface for MockSectionInterface {
    fn num_sections(&self) -> SecId {
        self.states.len() as SecId
    }
    fn set_section_state(&self, id: SecId, running: bool) {
        debug!(id, running, "setting section");
        self.states[id as usize].store(running, Ordering::SeqCst);
    }
    fn get_section_state(&self, id: SecId) -> bool {
        self.states[id as usize].load(Ordering::SeqCst)
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn test_mock_section_interface() {
        let iface = MockSectionInterface::new(6);
        assert_eq!(iface.num_sections(), 6);
        for i in 0..6u32 {
            assert_eq!(iface.get_section_state(i), false);
        }
        for i in 0..6u32 {
            iface.set_section_state(i, true);
        }
        for i in 0..6u32 {
            assert_eq!(iface.get_section_state(i), true);
        }
        for i in 0..6u32 {
            iface.set_section_state(i, false);
        }
        for i in 0..6u32 {
            assert_eq!(iface.get_section_state(i), false);
        }
    }
}
